// Untahris (Common Playground)
// copyright (C) 2006 longuens@users.sourceforge.net
// Released under GNU General Public License (see the file COPYING)
// --------------------------------------------------------------
// Header file for everything.

#include <vector>
#include <string>
#include <SDL/SDL.h>
#include <SDL/SDL_ttf.h>
#include <SDL/SDL_mixer.h>

using namespace std;

#define MAXX 128
#define MAXY 128

#define NEVER 2000000000

int ky[4] = {-1, 0, 1, 0};
int kx[4] = {0, 1, 0, -1};

typedef char int8;
typedef short int16;
typedef long int32;
typedef long long int64;

enum PlayerKey {
  keyUp,
  keyRight,
  keyDown,
  keyLeft,
  keyFire,
  keyEarth,
  keyAir,
  keyWater,
  keyJoinGame // fake key
  };

enum CellType {
  cFree,
  cBreakWall,
  cPermaWall,

  cSnakeHead,
  cSnakeDead,
  cSnakeLeft,
  cSnakeRight,
  cSnakeUp,
  cSnakeDown,
  cSnakeTRON,
  
  cTetris,
  cTetrisDown,
  
  cBig,
  
  cCannonF,
  cCannonA,

  cBonusLevel,
  cBonusScore,
  cBonusAmmo,
  
  cCastleWall,

  cBomb1,
  cBomb2,
  cExplosion
  };

struct TBuffer;

enum IconId {
  icFree, icBreak, icPerma,
  icAmmo, icLevel, icScore, 
  icThrust, icBall, icBullet, icBomber, icBomb1, icBomb2, icSpawn,
  icExplosion, icExplosion1, icExplosion2, icExplosion3, 
  icCastle, icCannon, icCannonA, icCannonP, icTarget,
  icWorm, icBazooka,
  icLastIcon
  };

// display's icon indexed by given ID
#define Ic(x) icons[int(x)]
#define DIc(x) ow()->icons[int(x)]

static const int numIcons = int(icLastIcon);

static char* iconnames[numIcons] = {
  "FREE", "BREAK", "PERMA",
  "AMMO", "LEVEL", "SCORE",
  "THRUST", "BULLET", "BALL", "BOMBER", "BOMB1", "BOMB2", "SPAWN",
  "EXPLOSION", "EXPLOSION1", "EXPLOSION2", "EXPLOSION3", 
  "CASTLE", "CANNON", "CANNONA", "CANNONP", "TARGET",
  "WORM", "BAZOOKA"
  };

struct TIcon {
  int size;
  int pixel[64][64];
  };

struct TDisplayIcon {
  TIcon i;
  void loadFrom(FILE *f);
  void fillWith(char *chars, int size);
  void fillTile(TDisplayIcon &orig, int size);
  void gradFill(int x0, int y0, int x1, int y1, int c0, int c1);
  void gradBall(int x0, int y0, int x1, int y1, int c0, int c1);
  };

struct TPlayerIcon {
  TIcon i;
  int color;
  void drawAt(int x, int y);
  void assignColor(int c);
  void drawAtCentered(int x, int y) { drawAt(x-i.size/2, y-i.size/2); }
  void drawAtRotated(int x, int y, int rx, int ry);
  };

struct TPlayer {
  TPlayerIcon icons[numIcons];
  int8 index;
  int8 typenum;
  int16 lives, frags;
  int32 score, level, ammo, typeval;
  int16 spx, spy; // spawn location
  char *getTypeName();
  int32 color; // color used
  bool key[8]; // keys pressed
  void reactToKey(PlayerKey pk, bool pressed);
  void draw();
  void localDraw() {}
  void init(int c);
  bool useAmmo(int qty) 
    {if(ammo < qty) return false; ammo -= qty; return true;}
  void save(TBuffer& b);
  void load(TBuffer& b);
  void createIcons();
  };

struct TPlayObject {
  public:
  int8 owner;
  int32 time;
  int getTime() {return time;}
  virtual ~TPlayObject () {}
  virtual void reactToKey(PlayerKey pk, bool pressed) {}
  virtual void draw() {}
  virtual void go() {}
  TPlayer *ow();

  int spd() { return 7999 / (200 + ow()->level) + 1; }
  int scoreGain() { return (100+ow()->level)*(100+ow()->level)/1000; }
  int bulletSpeed() { return 50 + ow()->level/4; }
  virtual void save(TBuffer& b) = 0;
  
  // these methods are supposed to check whether the two collide
  virtual bool reactToBullet(struct OBullet* bullet) {return false;}
  // return true if the bullet is to disappear
  virtual bool reactToPad(struct OPad* pad) {return false;}
  // return true if it kills the pad (reactToPad destroys pad object)
  virtual bool reactToThruster(struct OThruster* bullet) {return false;}
  // return true if it kills the thruster (react destroy thuster)
  virtual bool smashedByBall(struct OBall* ball) {return false;}
  // return true if it changes the ball's direction
  virtual void destroyBalls() { }
  // react to 'no balls' command
  virtual bool catchesRope(int rx, int ry) {return false;}
  // does ninja rope at (rx,ry) hook to it?
  virtual void pullRope(int& rx, int& ry, int &drx, int &dry, int ddx, int ddy) {}
  // pull the ninja rope at (x,y)
  
  void killedBy(int player);
  // player killed me -- this deletes the object and scores a frag
  };

struct TPlayground {
  int32 time;
  int32 kqpos;
  
  int64 seed;
  int rand();

  int16 sizeX, sizeY;
  CellType map[MAXY][MAXX];
  char owner[MAXY][MAXX];
  
  void mapSet(int x, int y, CellType c, char o) {
    map[y][x] = c;
    owner[y][x] = o;
    }    
    
  vector<TPlayer*> pl;
  vector<TPlayObject*> objects;

  void upTime(int newTime);
  
  void cleanup();
  void init(int sx, int sy); 
  void generate();
  
  int addPlayer(TPlayer* p);
  
  void deleteObject(TPlayObject *obj);

  void save(TBuffer& b);
  void load(TBuffer& b);
  };

struct TKeypress {
  int32 time;
  bool pressed;
  int32 player; // int32 because also used for color
  PlayerKey key;
  };

template<class T> static inline int size(vector<T> v) {return v.size();}

enum TGameType {gLocal, gHost, gClient};

enum TWorldType {
  wtEmpty,
  wtCavern,
  wtSwissCheese,
  wtConstruction,
  wtFrame,
  wtGrid,
  wtRuins,
  wtWall,
  wtLastWorldType
  };

static const int numWorldTypes = int(wtLastWorldType);

static char* worldNames[numWorldTypes] = {
  "EMPTY", "CAVERN", "SWISS CHEESE", "CONSTRUCTION", "FRAME", "GRID", "RUINS", "WALL"
  };

struct TConfig {
  TGameType gt;
  string server;
  int  port;
  int  res;
  int  sizeX;
  int  sizeY;
  int  resX;
  int  resY;
  int  initialWorld;
  int  initBonus;

  bool fs;
  bool centered;

  string err;
  bool ok;
  };

struct TLocalPlayer {
  int index;   // index in the player table
  int keys[8]; // keys for this player
  int color;
  };

struct TDisplay {

  TDisplayIcon icons[numIcons];
  
  Mix_Chunk 
    *sndBirth, *sndDeath, 
    *sndBlue, *sndGreen, *sndRed,
    *sndShoot, *sndHit, *sndPlace;

  SDL_Surface *surface;
  TTF_Font *font;
  int sizeX;
  int sizeY;
  int ofsX;
  int ofsY;
  int Res;
  bool stop;
  
  int *bground; // background image
  
  int interpretPixel(char ch);
  
  int delta; // time difference
  
  int getCurrentTime();
  void setCurrentTime(int val);

  void init();
  
  void initForGame();
  
  void loop();
  
  void draw();
  
  void putpix(int x, int y, int c);
  
  void playSoundAt(Mix_Chunk *sound, int x, int y);
  
  void addKeypress(int key, bool type);
  
  void putText(int x, int y, int color, const char *text, int align);
  
  void menuLoop();
  void menuKey(int key);
  void menuDraw();
  
  int menupos;

  void genBGround(int x0, int y0, int x1, int y1, int sy);
  };

struct TException {
  string msg;
  int data;
  string ds;
  TException(string m, int d=0, string s="") {msg=m; data=d; ds=s;}
  };

#define KQSIZE 32768

TConfig c;
TDisplay d;
TPlayground g;
vector<TLocalPlayer> lps;

TKeypress kqueue[KQSIZE];
int kqend;

void addKeyToQueue(TKeypress& kp);

void startGame();
void dataTransfer();
void broadcastKey(TKeypress& kp);

